/* Copyright (C) 2022 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergey Gorbunov, Sergei Zharko [committer] */

#include "KfMaterialMap.h"

#include "AlgoFairloggerCompat.h"
#include "KfSimd.h"

#include <cmath>
#include <iomanip>
#include <sstream>
#include <vector>

using cbm::algo::kf::fvec;
using cbm::algo::kf::MaterialMap;

// ---------------------------------------------------------------------------------------------------------------------
//
MaterialMap::MaterialMap(MaterialMap&& other) noexcept { this->Swap(other); }

// ---------------------------------------------------------------------------------------------------------------------
//
MaterialMap& MaterialMap::operator=(MaterialMap&& other) noexcept
{
  if (this != &other) {
    MaterialMap tmp(std::move(other));
    this->Swap(tmp);
  }
  return *this;
}

// ---------------------------------------------------------------------------------------------------------------------
//
void MaterialMap::Add(const MaterialMap& other, float zTarg)
{
  // The function allows to add a material layer either from the left or from the right to the station
  // NOTE: A symmetry of x-y is assumed
  constexpr int nRays    = 3;                   // Number of rays in a bin in a dimension
  const bool bRadialRays = !std::isnan(zTarg);  // Are rays radial (true, coming from target) or parallel (false)?
  const auto scaleFactor = bRadialRays ? ((other.fZref - zTarg) / (this->fZref - zTarg)) : 1.;
  const auto binSize     = 2. * scaleFactor * this->fXYmax / this->fNbins;  // Size of each bin dimension [cm]
  const auto stepSize    = binSize / static_cast<float>(nRays);             // Step between two neighboring rays [cm]

  // The coordinates of the first ray intersection with the other material layer [cm]
  float xBinOther = -this->fXYmax * scaleFactor + stepSize * 0.5;
  float yBinOther = xBinOther;

  // Loop over bins of the active (this)
  for (int iBinY = 0; iBinY < this->fNbins; ++iBinY) {
    for (int iBinX = 0; iBinX < this->fNbins; ++iBinX) {
      // Collect material using ray shooting
      float avgThickness = 0;  // Collected average thickness
      for (int iRayY = 0; iRayY < nRays; ++iRayY) {
        for (int iRayX = 0; iRayX < nRays; ++iRayX) {
          avgThickness += other.GetRadThick(xBinOther + iBinX * stepSize, yBinOther + iBinY * stepSize);
        }
      }
      this->fTable[iBinX + this->fNbins * iBinY] += avgThickness / (nRays * nRays);
      xBinOther += binSize;
    }
    yBinOther += binSize;
  }
  this->fZmin = std::min(this->fZmin, other.fZmin);
  this->fZmax = std::max(this->fZmax, other.fZmax);
}

// ---------------------------------------------------------------------------------------------------------------------
//
int MaterialMap::GetBin(float x, float y) const
{
  int i = static_cast<int>((x + fXYmax) * fFactor);
  int j = static_cast<int>((y + fXYmax) * fFactor);
  if (i < 0 || j < 0 || i >= fNbins || j >= fNbins) {
    return -1;
  }
  return i + j * fNbins;
}

// ---------------------------------------------------------------------------------------------------------------------
//
void MaterialMap::Initialize(int nBins, float xyMax, float zRef, float zMin, float zMax)
{
  fNbins = nBins;
  fXYmax = xyMax;
  fZref  = zRef;
  fZmin  = zMin;
  fZmax  = zMax;

  if (fNbins < 1) {
    std::stringstream aStream;
    aStream << "MaterialMap: object cannot be initialized with non-positive nBins = " << fNbins;
    throw std::logic_error(aStream.str());
  }

  if (fXYmax < 0.) {
    std::stringstream aStream;
    aStream << "MaterialMap: object cannot be initialized with non-positive XYmax = " << fXYmax << " [cm]";
    throw std::logic_error(aStream.str());
  }

  if (!((fZmin <= fZref) && (fZref <= zMax))) {
    std::stringstream aStream;
    aStream << "MaterialMap: object cannot be initialized with inconsistent Z: min " << fZmin << " ref " << fZref
            << " max " << fZmax << " [cm]";
    throw std::logic_error(aStream.str());
  }

  fFactor = 0.5 * fNbins / fXYmax;
  fTable.resize(fNbins * fNbins);
}

// ---------------------------------------------------------------------------------------------------------------------
//
void MaterialMap::Swap(MaterialMap& other) noexcept
{
  std::swap(fNbins, other.fNbins);
  std::swap(fXYmax, other.fXYmax);
  std::swap(fFactor, other.fFactor);
  std::swap(fZref, other.fZref);
  std::swap(fZmin, other.fZmin);
  std::swap(fZmax, other.fZmax);
  std::swap(fTable, other.fTable);  // Probably can cause segmentation violation (did not understand)
}

// ---------------------------------------------------------------------------------------------------------------------
//
std::string MaterialMap::ToString() const
{
  std::stringstream msg;
  using std::setw;
  msg << "[mat.map] nBins " << setw(6) << fNbins << ", XYmax " << setw(12) << fXYmax << " , z (ref, min, max) "
      << setw(12) << fZref << ' ' << setw(12) << fZmin << ' ' << setw(12) << fZmax;
  return msg.str();
}
