/* Copyright (C) 2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// @file   KfTarget.h
/// @brief  A target layer in the KF-setup (header)
/// @since  18.07.2024
/// @author Sergei Zharko <s.zharko@gsi.de>

#pragma once

#include "KfDefs.h"
#include "KfGeoLayer.h"

#include <boost/serialization/access.hpp>
#include <boost/serialization/base_object.hpp>

namespace cbm::algo::kf
{
  /// \class Target
  /// \brief A geometry layer in the target region
  class Target : public GeoLayer {
   public:
    /// \brief Default constructor
    Target() = default;

    /// \brief Constructor from parameters
    /// \param materialMap  Instance of the material budget map
    /// \param x            x-coordinate of the nominal target center [cm]
    /// \param y            y-coordinate of the nominal target center [cm]
    /// \param z            z-coordinate of the nominal target center [cm]
    Target(kf::MaterialMap&& materialMap, float x, float y, float z);

    /// \brief Copy constructor
    Target(const Target&) = default;

    /// \brief Move constructor
    Target(Target&&) = default;

    /// \brief Copy assignment operator
    Target& operator=(const Target&) = default;

    /// \brief Move assignment operator
    Target& operator=(Target&&) = default;

    /// \brief Destructor
    ~Target() = default;

    /// \brief Gets x-coordinate of the nominal target center
    float GetX() const { return fX; }

    /// \brief Gets x-coordinate of the nominal target center
    float GetY() const { return fY; }

    /// \brief Gets x-coordinate of the nominal target center
    float GetZ() const { return fZ; }

    /// \brief Sets x-coordinate of the nominal target center
    /// \param x  x-coordinate [cm]
    void SetX(float x) { fX = x; }

    /// \brief Sets y-coordinate of the nominal target center
    /// \param y  y-coordinate [cm]
    void SetY(float y) { fY = y; }

    /// \brief Sets x-coordinate of the nominal target center
    /// \param z  x-coordinate [cm]
    void SetZ(float z) { fZ = z; }

   private:
    /// \brief Serialization method
    friend class boost::serialization::access;
    template<class Archive>
    void serialize(Archive& ar, const unsigned int /*version*/)
    {
      ar& boost::serialization::base_object<GeoLayer>(*this);
      ar& fX;
      ar& fY;
      ar& fZ;
    }

    float fX = defs::Undef<float>;  ///< x-coordinate of the nominal target center [cm]
    float fY = defs::Undef<float>;  ///< y-coordinate of the nominal target center [cm]
    float fZ = defs::Undef<float>;  ///< z-coordinate of the nominal target center [cm]
  };
}  // namespace cbm::algo::kf
